@@ -16,24 +16,39 @@
 
 import abc
 import datetime
-import warnings
-
-from collections import namedtuple
+import enum
 from collections.abc import MutableMapping as _MutableMapping
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Mapping,
+    NamedTuple,
+    Optional,
+    Type,
+    TypeVar,
+    Union,
+    cast,
+)
 
-from bson.binary import (UuidRepresentation,
-                         ALL_UUID_REPRESENTATIONS,
-                         UUID_REPRESENTATION_NAMES)
+from bson.binary import (
+    ALL_UUID_REPRESENTATIONS,
+    UUID_REPRESENTATION_NAMES,
+    UuidRepresentation,
+)
 
-def _abstractproperty(func):
+
+def _abstractproperty(func: Callable[..., Any]) -> property:
     return property(abc.abstractmethod(func))
 
+
 _RAW_BSON_DOCUMENT_MARKER = 101
 
 
-def _raw_document_class(document_class):
+def _raw_document_class(document_class: Any) -> bool:
     """Determine if a document_class is a RawBSONDocument class."""
-    marker = getattr(document_class, '_type_marker', None)
+    marker = getattr(document_class, "_type_marker", None)
     return marker == _RAW_BSON_DOCUMENT_MARKER
 
 
@@ -46,13 +61,14 @@ class TypeEncoder(abc.ABC):
 
     See :ref:`custom-type-type-codec` documentation for an example.
     """
+
     @_abstractproperty
-    def python_type(self):
+    def python_type(self) -> Any:
         """The Python type to be converted into something serializable."""
         pass
 
     @abc.abstractmethod
-    def transform_python(self, value):
+    def transform_python(self, value: Any) -> Any:
         """Convert the given Python object into something serializable."""
         pass
 
@@ -66,13 +82,14 @@ class TypeDecoder(abc.ABC):
 
     See :ref:`custom-type-type-codec` documentation for an example.
     """
+
     @_abstractproperty
-    def bson_type(self):
+    def bson_type(self) -> Any:
         """The BSON type to be converted into our own type."""
         pass
 
     @abc.abstractmethod
-    def transform_bson(self, value):
+    def transform_bson(self, value: Any) -> Any:
         """Convert the given BSON value into our own type."""
         pass
 
@@ -89,9 +106,15 @@ class TypeCodec(TypeEncoder, TypeDecoder):
 
     See :ref:`custom-type-type-codec` documentation for an example.
     """
+
     pass
 
 
+_Codec = Union[TypeEncoder, TypeDecoder, TypeCodec]
+_Fallback = Callable[[Any], Any]
+_DocumentType = TypeVar("_DocumentType", bound=Mapping[str, Any])
+
+
 class TypeRegistry(object):
     """Encapsulates type codecs used in encoding and / or decoding BSON, as
     well as the fallback encoder. Type registries cannot be modified after
@@ -118,16 +141,20 @@ class TypeRegistry(object):
         :mod:`bson` can encode. See :ref:`fallback-encoder-callable`
         documentation for an example.
     """
-    def __init__(self, type_codecs=None, fallback_encoder=None):
+
+    def __init__(
+        self,
+        type_codecs: Optional[Iterable[_Codec]] = None,
+        fallback_encoder: Optional[_Fallback] = None,
+    ) -> None:
         self.__type_codecs = list(type_codecs or [])
         self._fallback_encoder = fallback_encoder
-        self._encoder_map = {}
-        self._decoder_map = {}
+        self._encoder_map: Dict[Any, Any] = {}
+        self._decoder_map: Dict[Any, Any] = {}
 
         if self._fallback_encoder is not None:
             if not callable(fallback_encoder):
-                raise TypeError("fallback_encoder %r is not a callable" % (
-                    fallback_encoder))
+                raise TypeError("fallback_encoder %r is not a callable" % (fallback_encoder))
 
         for codec in self.__type_codecs:
             is_valid_codec = False
@@ -140,39 +167,84 @@ class TypeRegistry(object):
                 self._decoder_map[codec.bson_type] = codec.transform_bson
             if not is_valid_codec:
                 raise TypeError(
-                    "Expected an instance of %s, %s, or %s, got %r instead" % (
-                        TypeEncoder.__name__, TypeDecoder.__name__,
-                        TypeCodec.__name__, codec))
+                    "Expected an instance of %s, %s, or %s, got %r instead"
+                    % (TypeEncoder.__name__, TypeDecoder.__name__, TypeCodec.__name__, codec)
+                )
 
-    def _validate_type_encoder(self, codec):
+    def _validate_type_encoder(self, codec: _Codec) -> None:
         from bson import _BUILT_IN_TYPES
+
         for pytype in _BUILT_IN_TYPES:
-            if issubclass(codec.python_type, pytype):
-                err_msg = ("TypeEncoders cannot change how built-in types are "
-                           "encoded (encoder %s transforms type %s)" %
-                           (codec, pytype))
+            if issubclass(cast(TypeCodec, codec).python_type, pytype):
+                err_msg = (
+                    "TypeEncoders cannot change how built-in types are "
+                    "encoded (encoder %s transforms type %s)" % (codec, pytype)
+                )
                 raise TypeError(err_msg)
 
     def __repr__(self):
-        return ('%s(type_codecs=%r, fallback_encoder=%r)' % (
-            self.__class__.__name__, self.__type_codecs,
-            self._fallback_encoder))
+        return "%s(type_codecs=%r, fallback_encoder=%r)" % (
+            self.__class__.__name__,
+            self.__type_codecs,
+            self._fallback_encoder,
+        )
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> Any:
         if not isinstance(other, type(self)):
             return NotImplemented
-        return ((self._decoder_map == other._decoder_map) and
-                (self._encoder_map == other._encoder_map) and
-                (self._fallback_encoder == other._fallback_encoder))
+        return (
+            (self._decoder_map == other._decoder_map)
+            and (self._encoder_map == other._encoder_map)
+            and (self._fallback_encoder == other._fallback_encoder)
+        )
 
 
-_options_base = namedtuple(
-    'CodecOptions',
-    ('document_class', 'tz_aware', 'uuid_representation',
-     'unicode_decode_error_handler', 'tzinfo', 'type_registry'))
+class DatetimeConversion(int, enum.Enum):
+    """Options for decoding BSON datetimes."""
+
+    DATETIME = 1
+    """Decode a BSON UTC datetime as a :class:`datetime.datetime`.
+
+    BSON UTC datetimes that cannot be represented as a
+    :class:`~datetime.datetime` will raise an :class:`OverflowError`
+    or a :class:`ValueError`.
+
+    .. versionadded 4.3
+    """
+
+    DATETIME_CLAMP = 2
+    """Decode a BSON UTC datetime as a :class:`datetime.datetime`, clamping
+    to :attr:`~datetime.datetime.min` and :attr:`~datetime.datetime.max`.
+
+    .. versionadded 4.3
+    """
+
+    DATETIME_MS = 3
+    """Decode a BSON UTC datetime as a :class:`~bson.datetime_ms.DatetimeMS`
+    object.
+
+    .. versionadded 4.3
+    """
+
+    DATETIME_AUTO = 4
+    """Decode a BSON UTC datetime as a :class:`datetime.datetime` if possible,
+    and a :class:`~bson.datetime_ms.DatetimeMS` if not.
+
+    .. versionadded 4.3
+    """
 
 
-class CodecOptions(_options_base):
+class _BaseCodecOptions(NamedTuple):
+    document_class: Type[Mapping[str, Any]]
+    tz_aware: bool
+    uuid_representation: int
+    unicode_decode_error_handler: str
+    tzinfo: Optional[datetime.tzinfo]
+    type_registry: TypeRegistry
+    datetime_conversion: Optional[DatetimeConversion]
+
+
+class CodecOptions(_BaseCodecOptions):
     """Encapsulates options used encoding and / or decoding BSON.
 
     The `document_class` option is used to define a custom type for use
@@ -207,7 +279,7 @@ class CodecOptions(_options_base):
     See :doc:`/examples/datetimes` for examples using the `tz_aware` and
     `tzinfo` options.
 
-    See :doc:`examples/uuid` for examples using the `uuid_representation`
+    See :doc:`/examples/uuid` for examples using the `uuid_representation`
     option.
 
     :Parameters:
@@ -226,13 +298,20 @@ class CodecOptions(_options_base):
       - `unicode_decode_error_handler`: The error handler to apply when
         a Unicode-related error occurs during BSON decoding that would
         otherwise raise :exc:`UnicodeDecodeError`. Valid options include
-        'strict', 'replace', and 'ignore'. Defaults to 'strict'.
+        'strict', 'replace', 'backslashreplace', 'surrogateescape', and
+        'ignore'. Defaults to 'strict'.
       - `tzinfo`: A :class:`~datetime.tzinfo` subclass that specifies the
         timezone to/from which :class:`~datetime.datetime` objects should be
         encoded/decoded.
       - `type_registry`: Instance of :class:`TypeRegistry` used to customize
         encoding and decoding behavior.
-
+      - `datetime_conversion`: Specifies how UTC datetimes should be decoded
+        within BSON. Valid options include 'datetime_ms' to return as a
+        DatetimeMS, 'datetime' to return as a datetime.datetime and
+        raising a ValueError for out-of-range values, 'datetime_auto' to
+        return DatetimeMS objects when the underlying datetime is
+        out-of-range and 'datetime_clamp' to clamp to the minimum and
+        maximum possible datetimes. Defaults to 'datetime'.
     .. versionchanged:: 4.0
        The default for `uuid_representation` was changed from
        :const:`~bson.binary.UuidRepresentation.PYTHON_LEGACY` to
@@ -248,31 +327,44 @@ class CodecOptions(_options_base):
        and stored back to the server.
     """
 
-    def __new__(cls, document_class=dict,
-                tz_aware=False,
-                uuid_representation=UuidRepresentation.UNSPECIFIED,
-                unicode_decode_error_handler="strict",
-                tzinfo=None, type_registry=None):
-        if not (issubclass(document_class, _MutableMapping) or
-                _raw_document_class(document_class)):
-            raise TypeError("document_class must be dict, bson.son.SON, "
-                            "bson.raw_bson.RawBSONDocument, or a "
-                            "sublass of collections.abc.MutableMapping")
+    def __new__(
+        cls: Type["CodecOptions"],
+        document_class: Optional[Type[Mapping[str, Any]]] = None,
+        tz_aware: bool = False,
+        uuid_representation: Optional[int] = UuidRepresentation.UNSPECIFIED,
+        unicode_decode_error_handler: str = "strict",
+        tzinfo: Optional[datetime.tzinfo] = None,
+        type_registry: Optional[TypeRegistry] = None,
+        datetime_conversion: Optional[DatetimeConversion] = DatetimeConversion.DATETIME,
+    ) -> "CodecOptions":
+        doc_class = document_class or dict
+        # issubclass can raise TypeError for generic aliases like SON[str, Any].
+        # In that case we can use the base class for the comparison.
+        is_mapping = False
+        try:
+            is_mapping = issubclass(doc_class, _MutableMapping)
+        except TypeError:
+            if hasattr(doc_class, "__origin__"):
+                is_mapping = issubclass(doc_class.__origin__, _MutableMapping)  # type: ignore[union-attr]
+        if not (is_mapping or _raw_document_class(doc_class)):
+            raise TypeError(
+                "document_class must be dict, bson.son.SON, "
+                "bson.raw_bson.RawBSONDocument, or a "
+                "subclass of collections.abc.MutableMapping"
+            )
         if not isinstance(tz_aware, bool):
             raise TypeError("tz_aware must be True or False")
         if uuid_representation not in ALL_UUID_REPRESENTATIONS:
-            raise ValueError("uuid_representation must be a value "
-                             "from bson.binary.UuidRepresentation")
-        if not isinstance(unicode_decode_error_handler, (str, None)):
-            raise ValueError("unicode_decode_error_handler must be a string "
-                             "or None")
+            raise ValueError(
+                "uuid_representation must be a value from bson.binary.UuidRepresentation"
+            )
+        if not isinstance(unicode_decode_error_handler, str):
+            raise ValueError("unicode_decode_error_handler must be a string")
         if tzinfo is not None:
             if not isinstance(tzinfo, datetime.tzinfo):
-                raise TypeError(
-                    "tzinfo must be an instance of datetime.tzinfo")
+                raise TypeError("tzinfo must be an instance of datetime.tzinfo")
             if not tz_aware:
-                raise ValueError(
-                    "cannot specify tzinfo without also setting tz_aware=True")
+                raise ValueError("cannot specify tzinfo without also setting tz_aware=True")
 
         type_registry = type_registry or TypeRegistry()
 
@@ -280,40 +372,58 @@ class CodecOptions(_options_base):
             raise TypeError("type_registry must be an instance of TypeRegistry")
 
         return tuple.__new__(
-            cls, (document_class, tz_aware, uuid_representation,
-                  unicode_decode_error_handler, tzinfo, type_registry))
+            cls,
+            (
+                doc_class,
+                tz_aware,
+                uuid_representation,
+                unicode_decode_error_handler,
+                tzinfo,
+                type_registry,
+                datetime_conversion,
+            ),
+        )
 
-    def _arguments_repr(self):
+    def _arguments_repr(self) -> str:
         """Representation of the arguments used to create this object."""
-        document_class_repr = (
-            'dict' if self.document_class is dict
-            else repr(self.document_class))
+        document_class_repr = "dict" if self.document_class is dict else repr(self.document_class)
 
-        uuid_rep_repr = UUID_REPRESENTATION_NAMES.get(self.uuid_representation,
-                                                      self.uuid_representation)
+        uuid_rep_repr = UUID_REPRESENTATION_NAMES.get(
+            self.uuid_representation, self.uuid_representation
+        )
 
-        return ('document_class=%s, tz_aware=%r, uuid_representation=%s, '
-                'unicode_decode_error_handler=%r, tzinfo=%r, '
-                'type_registry=%r' %
-                (document_class_repr, self.tz_aware, uuid_rep_repr,
-                 self.unicode_decode_error_handler, self.tzinfo,
-                 self.type_registry))
+        return (
+            "document_class=%s, tz_aware=%r, uuid_representation=%s, "
+            "unicode_decode_error_handler=%r, tzinfo=%r, "
+            "type_registry=%r, datetime_conversion=%s"
+            % (
+                document_class_repr,
+                self.tz_aware,
+                uuid_rep_repr,
+                self.unicode_decode_error_handler,
+                self.tzinfo,
+                self.type_registry,
+                self.datetime_conversion,
+            )
+        )
 
-    def _options_dict(self):
+    def _options_dict(self) -> Dict[str, Any]:
         """Dictionary of the arguments used to create this object."""
         # TODO: PYTHON-2442 use _asdict() instead
         return {
-            'document_class': self.document_class,
-            'tz_aware': self.tz_aware,
-            'uuid_representation': self.uuid_representation,
-            'unicode_decode_error_handler': self.unicode_decode_error_handler,
-            'tzinfo': self.tzinfo,
-            'type_registry': self.type_registry}
+            "document_class": self.document_class,
+            "tz_aware": self.tz_aware,
+            "uuid_representation": self.uuid_representation,
+            "unicode_decode_error_handler": self.unicode_decode_error_handler,
+            "tzinfo": self.tzinfo,
+            "type_registry": self.type_registry,
+            "datetime_conversion": self.datetime_conversion,
+        }
 
     def __repr__(self):
-        return '%s(%s)' % (self.__class__.__name__, self._arguments_repr())
+        return "%s(%s)" % (self.__class__.__name__, self._arguments_repr())
 
-    def with_options(self, **kwargs):
+    def with_options(self, **kwargs: Any) -> "CodecOptions":
         """Make a copy of this CodecOptions, overriding some options::
 
             >>> from bson.codec_options import DEFAULT_CODEC_OPTIONS
@@ -333,15 +443,20 @@ class CodecOptions(_options_base):
 DEFAULT_CODEC_OPTIONS = CodecOptions()
 
 
-def _parse_codec_options(options):
+def _parse_codec_options(options: Any) -> CodecOptions:
     """Parse BSON codec options."""
     kwargs = {}
-    for k in set(options) & {'document_class', 'tz_aware',
-                             'uuidrepresentation',
-                             'unicode_decode_error_handler', 'tzinfo',
-                             'type_registry'}:
-        if k == 'uuidrepresentation':
-            kwargs['uuid_representation'] = options[k]
+    for k in set(options) & {
+        "document_class",
+        "tz_aware",
+        "uuidrepresentation",
+        "unicode_decode_error_handler",
+        "tzinfo",
+        "type_registry",
+        "datetime_conversion",
+    }:
+        if k == "uuidrepresentation":
+            kwargs["uuid_representation"] = options[k]
         else:
             kwargs[k] = options[k]
     return CodecOptions(**kwargs)
